import asyncio
import time
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor
from logging import Logger
from pathlib import Path
from threading import Event
from typing import Callable, Optional, Union

import aiohttp

from .downloader import Multidown, Singledown
from .utls import (
    AutoShutdownFuture,
    FileValidator,
    ScreenCleaner,
    combine_files,
    create_segment_table,
    cursor_up,
    default_logger,
    get_filepath,
    seconds_to_hms,
    to_mb,
)


class Pypdl:
    """
    A multi-segment file downloader that supports progress tracking, retries, pause/resume functionality etc.

    This class also supports additional keyword arguments specified in the documentation.
    """

    def __init__(
        self,
        allow_reuse: bool = False,
        logger: Logger = default_logger("Pypdl"),
        **kwargs,
    ):
        self._pool = ThreadPoolExecutor(max_workers=2)
        self._workers = []
        self._interrupt = Event()
        self._stop = False
        self._kwargs = {
            "timeout": aiohttp.ClientTimeout(sock_read=60),
            "raise_for_status": True,
        }
        self._kwargs.update(kwargs)
        self._allow_reuse = allow_reuse

        self.size = None
        self.progress = 0
        self.speed = 0
        self.time_spent = 0
        self.current_size = 0
        self.eta = "99:59:59"
        self.remaining = None
        self.failed = False
        self.completed = False
        self.wait = True
        self.logger = logger

    def start(
        self,
        url: str,
        file_path: Optional[str] = None,
        multisegment: bool = True,
        segments: int = 10,
        overwrite: bool = True,
        etag: bool = True,
        retries: int = 0,
        mirror_func: Optional[Callable[[], str]] = None,
        display: bool = True,
        clear_terminal: bool = True,
        block: bool = True,
    ) -> Union[AutoShutdownFuture, Future, FileValidator, None]:
        """
        Start the download process.

        Parameters:

            url (str or Callable[[], str], required):
                The URL of the file to download, or a function that returns the URL.

            file_path (str, optional):
                The path to save the downloaded file.
                    If not provided, the file is saved in the current working directory with its original name.
                    If `file_path` is a directory, the file is saved in that directory with its original name.
                    If `file_path` is a file name, the file is saved with that name.

            multisegment (bool, optional):
                Whether to use multi-segment download. Default is `True`.

            segments (int, optional):
                The number of segments to divide the file into for multi-segment download. Default is `10`.

            overwrite (bool, optional):
                Whether to overwrite the file if it already exists. Default is `True`.

            etag (bool, optional):
                Whether to validate the ETag before resuming downloads. Default is `True`.

            retries (int, optional):
                The number of times to retry the download if it fails. Default is `0`.

            mirror_func (Callable[[], str], optional):
                A function that returns a new download URL if the download fails. Default is `None`.

            display (bool, optional):
                Whether to display download progress and other messages. Default is `True`.

            clear_terminal (bool, optional):
                Whether to clear the terminal before displaying the download progress. Default is `True`.

            block (bool, optional):
                Whether to block the function until the download is complete. Default is `True`.

        Returns:
            AutoShutdownFuture: If `block` is `False`.
            concurrent.futures.Future: If `block` is `False` and `allow_reuse` is `True`.
            FileValidator: If `block` is `True` and the download is successful.
            None: If `block` is `True` and the download fails.
        """

        def download():
            for i in range(retries + 1):
                try:
                    _url = mirror_func() if i > 0 and callable(mirror_func) else url
                    self.logger.debug("Downloading, url: %s attempt: %s", _url, (i + 1))
                    result = self._execute(
                        _url,
                        file_path,
                        multisegment,
                        segments,
                        overwrite,
                        etag,
                        display,
                        clear_terminal,
                    )

                    if self._stop or self.completed:
                        if display:
                            print(f"Time elapsed: {seconds_to_hms(self.time_spent)}")
                        return result

                    self._reset()
                    time.sleep(3)

                except Exception as e:
                    self.logger.exception("(%s) [%s]", e.__class__.__name__, e)

            self.wait = False
            self.failed = True
            self.logger.debug("Download failed, url: %s", _url)
            return None

        self._reset()
        url = url() if callable(url) else url

        if self._allow_reuse:
            future = self._pool.submit(download)
        else:
            future = AutoShutdownFuture(self._pool.submit(download), [self._pool])

        if block:
            result = future.result()
            return result

        return future

    def stop(self) -> None:
        """Stop the download process."""
        self._interrupt.set()
        self._stop = True
        time.sleep(1)
        self.logger.debug("Download stopped")

    def shutdown(self) -> None:
        """Shutdown the download manager."""
        self._pool.shutdown()
        self.logger.debug("Shutdown download manager")

    def _reset(self):
        self._workers.clear()
        self._interrupt.clear()
        self._stop = False

        self.size = None
        self.progress = 0
        self.speed = 0
        self.time_spent = 0
        self.current_size = 0
        self.eta = "99:59:59"
        self.remaining = None
        self.failed = False
        self.completed = False
        self.wait = True
        self.logger.debug("Reset download manager")

    def _execute(
        self,
        url,
        file_path,
        multisegment,
        segments,
        overwrite,
        etag,
        display,
        clear_terminal,
    ):
        start_time = time.time()

        file_path, multisegment, etag = self._get_info(
            url, file_path, multisegment, etag
        )

        if not overwrite and Path(file_path).exists():
            self.wait = False
            self.completed = True
            self.time_spent = time.time() - start_time
            self.logger.debug("File already exists, download completed")
            return FileValidator(file_path)

        if multisegment:
            segment_table = create_segment_table(
                url, file_path, segments, self.size, etag
            )
            self.logger.debug("Segment table created: %s", str(segment_table))
            segments = segment_table["segments"]

            self._pool.submit(
                lambda: asyncio.run(self._multi_segment(segments, segment_table))
            )
        else:
            self._pool.submit(lambda: asyncio.run(self._single_segment(url, file_path)))

        recent_queue = deque([0] * 12, maxlen=12)
        download_mode = "Multi-Segment" if multisegment else "Single-Segment"
        interval = 0.5
        self.wait = False
        self.logger.debug("Initiated waiting loop")
        with ScreenCleaner(display, clear_terminal):
            while True:
                status = sum(worker.completed for worker in self._workers)
                self._calc_values(recent_queue, interval)

                if display:
                    self._display(download_mode)

                if self._interrupt.is_set():
                    self.time_spent = time.time() - start_time
                    self.logger.debug("Exit waiting loop, download interrupted")
                    return None

                if status and status == len(self._workers):
                    if multisegment:
                        self.logger.debug("Combining files")
                        combine_files(file_path, segments)
                    self.completed = True
                    self.time_spent = time.time() - start_time
                    self.logger.debug("Exit waiting loop, download completed")
                    return FileValidator(file_path)

                time.sleep(interval)

    def _get_info(self, url, file_path, multisegment, etag):
        header = asyncio.run(self._get_header(url))
        file_path = get_filepath(url, header, file_path)
        if size := int(header.get("content-length", 0)):
            self.logger.debug("Size acquired from header")
            self.size = size

        etag = header.get("etag", not etag)  # since we check truthiness of etag

        if isinstance(etag, str):
            self.logger.debug("ETag acquired from header")
            etag = etag.strip('"')

        if not self.size or not header.get("accept-ranges"):
            self.logger.debug("Single segment download, accept-ranges header not found")
            multisegment = False

        return file_path, multisegment, etag

    async def _get_header(self, url):
        async with aiohttp.ClientSession() as session:
            async with session.head(url, **self._kwargs) as response:
                if response.status == 200:
                    self.logger.debug("Header acquired from head request")
                    return response.headers

            async with session.get(url, **self._kwargs) as response:
                if response.status == 200:
                    self.logger.debug("Header acquired from get request")
                    return response.headers

        raise Exception(
            f"Failed to get header (Status: {response.status}, Reason: {response.reason})"
        )

    async def _multi_segment(self, segments, segment_table):
        tasks = []
        self.logger.debug("Multi-Segment download started")
        async with aiohttp.ClientSession() as session:
            for segment in range(segments):
                md = Multidown(self._interrupt)
                self._workers.append(md)
                tasks.append(
                    asyncio.create_task(
                        md.worker(segment_table, segment, session, **self._kwargs)
                    )
                )
            try:
                await asyncio.gather(*tasks)
                self.logger.debug("Downloaded all segments")
            except Exception as e:
                self.logger.exception("(%s) [%s]", e.__class__.__name__, e)
                self._interrupt.set()

    async def _single_segment(self, url, file_path):
        self.logger.debug("Single-Segment download started")
        async with aiohttp.ClientSession() as session:
            sd = Singledown(self._interrupt)
            self._workers.append(sd)
            try:
                await sd.worker(url, file_path, session, **self._kwargs)
                self.logger.debug("Downloaded single segment")
            except Exception as e:
                self.logger.exception("(%s) [%s]", e.__class__.__name__, e)
                self._interrupt.set()

    def _calc_values(self, recent_queue, interval):
        self.current_size = sum(worker.curr for worker in self._workers)

        # Speed calculation
        recent_queue.append(sum(worker.downloaded for worker in self._workers))
        non_zero_list = [to_mb(value) for value in recent_queue if value]
        if len(non_zero_list) < 1:
            self.speed = 0
        elif len(non_zero_list) == 1:
            self.speed = non_zero_list[0] / interval
        else:
            diff = [b - a for a, b in zip(non_zero_list, non_zero_list[1:])]
            self.speed = (sum(diff) / len(diff)) / interval

        if self.size:
            self.progress = int((self.current_size / self.size) * 100)
            self.remaining = to_mb(self.size - self.current_size)

            if self.speed:
                self.eta = seconds_to_hms(self.remaining / self.speed)
            else:
                self.eta = "99:59:59"

    def _display(self, download_mode):
        cursor_up()
        if self.size:
            progress_bar = f"[{'█' * self.progress}{'·' * (100 - self.progress)}] {self.progress}% \n"
            info = f"Total: {to_mb(self.size):.2f} MB, Download Mode: {download_mode}, Speed: {self.speed:.2f} MB/s, ETA: {self.eta} "
            print(progress_bar + info)
        else:
            download_stats = "Downloading... \n"
            info = f"Downloaded: {to_mb(self.current_size):.2f} MB, Download Mode: {download_mode}, Speed: {self.speed:.2f} MB/s "
            print(download_stats + info)



from pathlib import Path
from threading import Event

import aiofiles
from aiohttp import ClientSession

MEGABYTE = 1048576


class Basicdown:
    """Base downloader class."""

    def __init__(self, interrupt: Event):
        self.curr = 0  # Downloaded size in bytes (current size)
        self.completed = False
        self.interrupt = interrupt
        self.downloaded = 0

    async def download(
        self, url: str, path: str, mode: str, session: ClientSession, **kwargs
    ) -> None:
        """Download data in chunks."""
        async with session.get(url, **kwargs) as response:
            async with aiofiles.open(path, mode) as file:
                async for chunk in response.content.iter_chunked(MEGABYTE):
                    await file.write(chunk)
                    self.curr += len(chunk)
                    self.downloaded += len(chunk)
                    if self.interrupt.is_set():
                        break


class Singledown(Basicdown):
    """Class for downloading the whole file in a single segment."""

    async def worker(
        self, url: str, file_path: str, session: ClientSession, **kwargs
    ) -> None:
        await self.download(url, file_path, "wb", session, **kwargs)
        self.completed = True


class Multidown(Basicdown):
    """Class for downloading a specific segment of the file."""

    async def worker(
        self, segment_table: dict, id: int, session: ClientSession, **kwargs
    ) -> None:
        url = segment_table["url"]
        overwrite = segment_table["overwrite"]
        segment_path = Path(segment_table[id]["segment_path"])
        start = segment_table[id]["start"]
        end = segment_table[id]["end"]
        size = segment_table[id]["segment_size"]

        if segment_path.exists():
            downloaded_size = segment_path.stat().st_size
            if overwrite or downloaded_size > size:
                segment_path.unlink()
            else:
                self.curr = downloaded_size

        if kwargs.get("headers") is not None:
            kwargs["headers"] = kwargs["headers"].copy()

        if self.curr < size:
            start = start + self.curr
            kwargs.setdefault("headers", {}).update({"range": f"bytes={start}-{end}"})
            await self.download(url, segment_path, "ab", session, **kwargs)

        if self.curr == size:
            self.completed = True
        else:
            if not self.interrupt.is_set():
                raise Exception(
                    f"Incorrect segment size: expected {size} bytes, received {self.curr} bytes"
                )



import threading
import time
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from logging import Logger, getLogger
from typing import Union

from .pypdl_manager import Pypdl
from .utls import (
    AutoShutdownFuture,
    ScreenCleaner,
    cursor_up,
    default_logger,
    seconds_to_hms,
    to_mb,
)


class PypdlFactory:
    """
    A factory class for managing multiple instances of the Pypdl downloader.

    This class also supports additional keyword arguments specified in the documentation.
    """

    def __init__(
        self,
        instances: int = 2,
        allow_reuse: bool = False,
        logger: Logger = default_logger("PypdlFactory"),
        **kwargs,
    ):
        self._instances = [
            Pypdl(True, getLogger(f"PypdlFactory.instance-{i}"), **kwargs)
            for i in range(instances)
        ]
        self._allow_reuse = allow_reuse
        self._pool = ThreadPoolExecutor(max_workers=2)
        self._stop = False
        self._prog = True
        self._completed_size = 0
        self._completed_prog = 0
        self._lock = threading.Event()
        self._stop_lock = threading.Lock()
        self._running = []

        self.progress = 0
        self.speed = 0
        self.time_spent = 0
        self.current_size = 0
        self.total = 0
        self.success = []
        self.failed = []
        self.remaining = []
        self.logger = logger

    @property
    def completed(self) -> bool:
        """Check if all tasks are completed."""
        if self.total == 0:
            return False
        return len(self.success) + len(self.failed) == self.total

    def start(
        self,
        tasks: list,
        display: bool = True,
        clear_terminal: bool = True,
        block: bool = True,
    ) -> Union[AutoShutdownFuture, Future, list]:
        """
        Start the download process.

        Parameters:

            tasks (list):
                A list of tasks to be downloaded.
                    Each task is a tuple where the 1st element is the URL and the 2nd element is an optional dict with kwargs for pypdl start method.

            display (bool, Optional):
                Whether to display download progress and other messages. Default is True.

            clear_terminal  (bool, Optional):
                Whether to clear the terminal before displaying the download progress. Default is True.

            block (bool, Optional):
                Whether to block the function until all downloads are complete. Default is True.

        Returns:
            AutoShutdownFuture: If `block` is `False`.
            concurrent.futures.Future: If `block` is `False` and `allow_reuse` is `True`.
            list: If `block` is `True`, a list of (url, `FileValidator`) tuples for successfully completed tasks.
        """
        self._reset()
        self.logger.debug("Downloading %s files", len(tasks))
        if self._allow_reuse:
            future = self._pool.submit(self._execute, tasks, display, clear_terminal)
        else:
            future = AutoShutdownFuture(
                self._pool.submit(self._execute, tasks, display, clear_terminal),
                [*self._instances, self._pool],
            )

        if block:
            result = future.result()
            return result

        return future

    def stop(self) -> None:
        """Stops all active downloads."""
        if not self._running:
            return
        with self._stop_lock:
            self.logger.debug("Initiating download stop")
            self._lock.set()
            self._stop = True
        for instance in self._instances:
            instance.stop()
        while self._lock.is_set():
            time.sleep(0.5)
        time.sleep(1)
        self.logger.debug("Download stopped")

    def shutdown(self) -> None:
        """Shutdown the factory."""
        self.logger.debug("Shutting down factory")
        for instance in self._instances:
            instance.shutdown()
        self._pool.shutdown()
        self.logger.debug("Factory shutdown")

    def _reset(self):
        self._stop = False
        self._prog = True
        self._completed_size = 0
        self._completed_prog = 0
        self._running.clear()
        self.success.clear()
        self.failed.clear()
        self.remaining.clear()
        self.logger.debug("Reseted download factory")

    def _execute(self, tasks, display, clear_terminal):
        start_time = time.time()
        self.total = len(tasks)
        self.remaining = tasks[len(self._instances) :]
        futures = {}

        for instance, task in zip(self._instances, tasks):
            self._add_future(instance, task, futures)
            self._running.append(instance)

        self._pool.submit(self._compute, display, clear_terminal)

        self.logger.debug("Initiated waiting loop")
        while not self.completed:
            if self._stop:
                self.logger.debug("Exit waiting loop, download interrupted")
                break

            for future in as_completed(futures):
                instance, curr_url = futures.pop(future)

                if instance.completed:
                    self._handle_success(instance, curr_url, future.result())
                elif instance.failed:
                    self._handle_failed(curr_url)

                self._manage_remaining(instance, futures)

        self.time_spent = time.time() - start_time
        self.logger.debug("Completed, clearing lock")
        self._lock.clear()
        self.logger.debug("Exit waiting loop, download completed")
        return self.success

    def _add_future(self, instance, task, futures):
        self.logger.debug("Adding new task")
        url, *kwargs = task
        kwargs = kwargs[0] if kwargs else {}
        kwargs.update({"block": False, "display": False, "overwrite": False})
        future = instance.start(url, **kwargs)
        futures[future] = (instance, url)
        while instance.wait:
            time.sleep(0.1)
        self.logger.debug("Added new task: %s", url)

    def _handle_success(self, instance, curr_url, result):
        self.logger.debug("Handling completed download, setting lock")
        self._lock.set()
        if instance.size:
            self._completed_size += instance.size
        else:
            self._prog = False
        self._completed_prog += (1 / self.total) * 100
        self.success.append((curr_url, result))
        self.logger.debug("Download completed: %s", curr_url)

    def _handle_failed(self, curr_url):
        self.logger.debug("Handling failed download, setting lock")
        self._lock.set()
        self.failed.append(curr_url)
        self.logger.error("Download failed: %s", curr_url)

    def _manage_remaining(self, instance, futures):
        with self._stop_lock:
            if self._stop:
                self.logger.debug("Stop Initiated, removing instance from running")
                self._running.remove(instance)
                return

            if self.remaining:
                self.logger.debug("Remaining tasks: %s", len(self.remaining))
                self._add_future(instance, self.remaining.pop(0), futures)
            else:
                self.logger.debug("No remaining tasks, removing instance from running")
                self._running.remove(instance)

            if not self.completed:
                self.logger.debug("Not completed, releasing lock")
                self._lock.clear()
                time.sleep(0.5)

    def _compute(self, display, clear_terminal):
        self.logger.debug("Starting download computation")
        with ScreenCleaner(display, clear_terminal):
            while True:
                if not self._lock.is_set():
                    self._calc_values()

                    if display:
                        self._display()

                    if self._stop or self.completed:
                        break
                time.sleep(0.5)

            self.progress = round(self._completed_prog)
            self.current_size = self._completed_size
            if display:
                self._display()
                print(f"Time elapsed: {seconds_to_hms(self.time_spent)}")
            self.logger.debug("Computation ended")

    def _calc_values(self):
        def sum_attribute(instances, attribute):
            return sum(getattr(instance, attribute) for instance in instances)

        def average_attribute(instances, attribute, total):
            return sum_attribute(instances, attribute) // total

        if self._running:
            self.speed = average_attribute(self._running, "speed", len(self._running))

            self.progress = average_attribute(
                self._running, "progress", self.total
            ) + int(self._completed_prog)
            self.current_size = (
                sum_attribute(self._running, "current_size") + self._completed_size
            )

    def _display(self):
        cursor_up()
        completed = len(self.success)
        prog = (
            all(instance.size for instance in self._instances) if self._prog else False
        )

        if prog:
            progress_bar = f"[{'█' * self.progress}{'·' * (100 - self.progress)}] {self.progress}% \n"
            info = f"Total: {completed}/{self.total}, Downloaded: {to_mb(self.current_size):.2f} MB, Speed: {self.speed:.2f} MB/s "
            print(progress_bar + info)
        else:
            download_stats = f"Downloading...{' ' * 95}\n"
            info = f"Total: {completed}/{self.total}, Downloaded: {to_mb(self.current_size):.2f} MB, Speed: {self.speed:.2f} MB/s "
            print(download_stats + info)



import hashlib
import json
import logging
import sys
import time
from concurrent.futures import Executor, Future
from pathlib import Path
from typing import Dict, List, Union
from urllib.parse import unquote, urlparse

MEGABYTE = 1048576
BLOCKSIZE = 4096
BLOCKS = 1024
CHUNKSIZE = BLOCKSIZE * BLOCKS


def to_mb(size_in_bytes: int) -> float:
    return size_in_bytes / MEGABYTE


def seconds_to_hms(sec: float) -> str:
    time_struct = time.gmtime(sec)
    return time.strftime("%H:%M:%S", time_struct)


def cursor_up() -> None:
    sys.stdout.write("\x1b[1A" * 2)  # Move cursor up two lines
    sys.stdout.flush()


def get_filepath(url: str, headers: Dict, file_path: str) -> str:
    content_disposition = headers.get("Content-Disposition", None)

    if content_disposition and "filename=" in content_disposition:
        filename_start = content_disposition.index("filename=") + len("filename=")
        filename = content_disposition[filename_start:]  # Get name from headers
        filename = unquote(filename.strip('"'))  # Decode URL encodings
    else:
        filename = unquote(urlparse(url).path.split("/")[-1])  # Generate name from URL

    filename = filename.replace("/", "_")

    if file_path:
        file_path = Path(file_path)
        if file_path.is_dir():
            return str(file_path / filename)
        return str(file_path)

    return filename


def create_segment_table(
    url: str, file_path: str, segments: str, size: int, etag: Union[str, bool]
) -> Dict:
    """Create a segment table for multi-segment download."""
    progress_file = Path(file_path + ".json")
    overwrite = True

    if progress_file.exists():
        progress = json.loads(progress_file.read_text())

        if (etag is True) or (
            progress["etag"] and (progress["url"] == url and progress["etag"] == etag)
        ):
            segments = progress["segments"]
            overwrite = False

    progress_file.write_text(
        json.dumps(
            {"url": url, "etag": etag, "segments": segments},
            indent=4,
        )
    )

    dic = {"url": url, "segments": segments, "overwrite": overwrite}
    partition_size, add_bytes = divmod(size, segments)

    for segment in range(segments):
        start = partition_size * segment
        end = partition_size * (segment + 1) - 1  # since range is inclusive

        if segment == segments - 1:
            end += add_bytes

        segment_size = end - start + 1  # since range is inclusive
        dic[segment] = {
            "start": start,
            "end": end,
            "segment_size": segment_size,
            "segment_path": f"{file_path}.{segment}",
        }

    return dic


def combine_files(file_path: str, segments: int) -> None:
    """Combine the downloaded file segments into a single file."""
    with open(file_path, "wb") as dest:
        for segment in range(segments):
            segment_file = f"{file_path}.{segment}"
            with open(segment_file, "rb") as src:
                while True:
                    chunk = src.read(CHUNKSIZE)
                    if chunk:
                        dest.write(chunk)
                    else:
                        break

            Path(segment_file).unlink()

    progress_file = Path(f"{file_path}.json")
    progress_file.unlink()


def default_logger(name: str) -> logging.Logger:
    """Creates a default debugging logger."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.ERROR)
    handler = logging.FileHandler("pypdl.log", mode="a", delay=True)
    handler.setFormatter(
        logging.Formatter(
            "(%(name)s)  %(asctime)s - %(levelname)s: %(message)s",
            datefmt="%d-%m-%y %H:%M:%S",
        )
    )
    logger.addHandler(handler)
    return logger


class FileValidator:
    """A class used to validate the integrity of the file."""

    def __init__(self, path: str):
        self.path = path

    def calculate_hash(self, algorithm: str, **kwargs) -> str:
        hash_obj = hashlib.new(algorithm, **kwargs)
        with open(self.path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_obj.update(chunk)
        return hash_obj.hexdigest()

    def validate_hash(self, correct_hash: str, algorithm: str, **kwargs) -> bool:
        file_hash = self.calculate_hash(algorithm, **kwargs)
        return file_hash == correct_hash


class AutoShutdownFuture:
    """A Future object wrapper that shuts down the executors when the result is retrieved."""

    def __init__(self, future: Future, executors: List[Executor]):
        self.future = future
        self.executors = executors

    def result(self, timeout: float = None) -> Union[FileValidator, None]:
        result = self.future.result(timeout)
        for executor in self.executors:
            executor.shutdown()
        return result


class ScreenCleaner:
    """A context manager to clear the screen and hide cursor."""

    def __init__(self, display: bool, clear_screen: bool):
        self.display = display
        self.clear_screen = clear_screen

    def clear(self) -> None:
        sys.stdout.write(2 * "\n")
        if self.clear_screen:
            sys.stdout.write("\033c")  # Clear screen
        sys.stdout.write("\x1b[?25l")  # Hide cursor

    def __enter__(self):
        if self.display:
            self.clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.display:
            sys.stdout.write("\x1b[?25h")  # Show cursor
